(**********************)
(*  TP 8 : Fractales  *)
(**********************)

#open "graphics";;
type point = { x : float ; y : float };;

let ignore _ = ();;

let show (draw : int -> unit) depth =
  open_graph "";
  for i = 0 to depth do
    draw i;
    ignore (read_key ());
    clear_graph ()
  done;
  close_graph ();;

let show_ifs (draw : 'a vect -> int -> unit) ifs iter =
  open_graph "";
  ignore (read_key ());
  draw ifs iter;
  ignore (read_key ());
  close_graph ();;



(*---------------------------*)
(*  Gomtrie en Caml Light  *)
(*---------------------------*)

(* 1 *)

(* 2 *)

(* 3 *)


(*
  minus : point -> point
  translation : point -> point -> point
  symmetry : point -> point -> point
  rotation_origin : float -> point -> point
  rotation : point -> float -> point -> point
  homothety : float -> point -> point
 *)

(* 4 *)

(* 5 *)

(*
  draw_line : point -> point -> unit
  plot_pt : point -> unit
*)

(*---------------------------*)
(*  Construction par tapes  *)
(*---------------------------*)


(* La courbe de von Koch *)


(* 6 *)

(* koch_transformation : point -> point -> point vect *)


(* 7 *)

(* koch_draw : point -> point -> int -> unit *)




(* Autres factales *)

let cross a b =
  let m = translation (minus a) b in
  let c = (translation a o homothety 0.47) m in
  let milieu = homothety 0.5 m in
  let d = (translation a o translation milieu o rotation_origin (pi/.2.)
             o homothety 0.95) milieu in
  let e = (translation a o homothety 0.53) m in
    [|a; c; d; e; b|];;

let sierpinski a b =
  let m = ((homothety 0.5) o (translation (minus a))) b in
  let mil = translation m a in
  let pi_3 = pi /. 3. in
  let pt_2 = rotation a pi_3 mil in
  let pt_3 = translation m pt_2 in
    [|a; pt_2; pt_3; b|],[|true; false; true|];;

let dragon a b =
  let c = (translation a o rotation_origin (pi/.4.)
              o homothety (sqrt 2. /.2.) o translation (minus a)) b in
    [|a; c; b|],[|false; true|];;

(* 8 *)

(* draw : (point -> point -> point vect) -> point -> point -> int -> unit *)


(* 9 *)

(* bdraw :
 (point -> point -> point vect * bool vect) -> point -> point -> bool -> int -> unit *)


(*----------------------------*)
(*  Iterated Function System  *)
(*----------------------------*)


let darken increment p =
  let col = point_color (round p.x) (round p.y) in
  let blue = col mod 256 in
  let green = (col / 256) mod 256 in
  let red = (col / (256 * 256)) mod 256 in
    set_color (rgb
                 (max 0 (red - increment))
                 (max 0 (green - increment))
                 (max 0 (blue - increment)));
    plot_pt p;;


let conjugate h h_inv ifs =
  map_vect (fun f -> h o f o h_inv) ifs;;

let translate x_offset y_offset = (* translation de fractale *)
  let t = {x = x_offset; y = y_offset} in
    conjugate (translation t) (translation (minus t));;

let proba_conjugate h h_inv ifs =
  map_vect (fun (f, p) -> (h o f o h_inv, p)) ifs;;

let proba_translate x_offset y_offset = (* translation d'ifs avec proba *)
  let t = {x = x_offset; y = y_offset} in
    proba_conjugate (translation t) (translation (minus t));;


let square_ifs width height =
  let size = 0.9 *. float_of_int (min width height) in
  let f = homothety 0.45 in
  let side = 0.55 *. size in
    [| f;
       translation {x = side; y = 0.} o f;
       translation {x = side; y = side} o f;
       translation {x = 0.; y = side} o f|];;

let sierpinski_ifs width height =  (* Triangle de Sierpinski *)
  let size = min (float_of_int width) (float_of_int (height * 2) /. sqrt 3.7) in
  let f = homothety 0.5 in
    [| f;
       translation {x = 0.50 *. size; y = 0.} o f;
       translation {x = 0.25 *. size; y = size *. sqrt 3. /. 4.} o f|];;

let maple_ifs width height = (* Feuille d'rable *)
  let size = float_of_int (min width height) in
  let h = homothety 0.5 in
  let r1 = rotation_origin (pi /. 4.) in
  let r2 = rotation_origin (-. pi /. 4.) in
  let c = { x = size /. 2. ; y = size /. 20. } in
    [|(translation {x = 0.1   *. size; y =  0.04 *. size}) o (homothety 0.8);
      (translation {x = 0.25  *. size; y =  0.40 *. size}) o h;
      (translation {x = 0.266 *. size; y = 0.078 *. size}) o h o r1;
      (translation {x = 0.378 *. size; y = 0.434 *. size}) o h o r2|];;

let dragon_ifs width height = (* Dragons jumeaux *)
  let size = float_of_int (min width height) in
  let f = (homothety (1. /. (sqrt 2.))) o (rotation_origin (-. pi /. 4.)) in
    [|(translation {x =    0.125 *. size ; y = 0.625 *. size}) o f;
      (translation {x = -. 0.125 *. size ; y = 0.375 *. size}) o f|];;

let pentagon_ifs width height = (* Pentagone de Sierpinski *)
  let size = 0.5 *. float_of_int (min width height) in
  let h = homothety ((3. -. sqrt 5. ) /. 2.) in
    [| h;
       translation { x =    0.618 *. size ; y = 0. } o h;
       translation { x =    0.809 *. size ; y = 0.588 *. size } o h;
       translation { x =    0.309 *. size ; y = 0.951 *. size } o h;
       translation { x = -. 0.191 *. size ; y = 0.588 *. size } o h|];;


let fern_proba width height = (* Fougre *)
  let size = 0.25 *. float_of_int (min height (width / 5)) in
  let x_offset = size /. 2. in
  let y_offset = 20. in
  let center = translation {x = 0. ; y = 0. } in
  [|(fun p -> {x =    0.85 *. p.x +. 0.04 *. p.y;
               y = -. 0.04 *. p.x +. 0.85 *. p.y +. 1.60 *. size}), 0.85;
    (fun p -> {x =    0.20 *. p.x -. 0.26 *. p.y;
	       y =    0.23 *. p.x +. 0.22 *. p.y +. 0.16 *. size}), 0.07;
    (fun p -> {x = -. 0.15 *. p.x +. 0.28 *. p.y;
	       y =    0.26 *. p.x +. 0.24 *. p.y +. 0.44 *. size}), 0.07;
    (fun p -> {x = 0.; y = 0.16 *. p.y}), 0.01|];;
  
let maple_proba width height = (* Nervures de la feuille d'rable *)
  let size = float_of_int (min width height) in
  let h = homothety 0.5 in
  let pi_4 = pi /. 4. in
  let c = { x = size /. 2. ; y = size /. 20. } in
    [|(translation {x = 0.1   *. size; y = 0.04  *. size})o(homothety 0.8),0.70;
      (translation {x = 0.25  *. size; y = 0.4   *. size}) o h, 0.10;
      (translation {x = 0.266 *. size; y = 0.078 *. size})
	o h o (rotation_origin pi_4), 0.10;
      (translation {x = 0.378 *. size; y = 0.434 *. size})
	o h o (rotation_origin (-. pi_4)), 0.10|];;

(* 10 *)

(* new_point : (point -> point) vect -> point -> point *)

(* 11 *)

(* draw_ifs : (point -> point) vect -> int -> unit *)

(* 13 *)

(* proba_ifs : float -> ((point -> point) * float) vect -> int -> unit *)

(* BONUS: test de tous les ifs *)
let full_test_ifs width height =
  let f_width = float_of_int width in
  let test_ifs draw ifs iter =
    clear_graph ();
    draw ifs iter;
    ignore (read_key ())
  in
    open_graph (string_of_int width ^ "x" ^string_of_int height);
    ignore (read_key ());
    test_ifs draw_ifs (square_ifs width height) 100000;
    test_ifs draw_ifs (sierpinski_ifs width height) 100000;
    test_ifs draw_ifs (koch_ifs width height) 50000;
    test_ifs draw_ifs (dragon_ifs width height) 150000;
    test_ifs draw_ifs (maple_ifs width height) 150000;
    test_ifs draw_ifs (translate 300. 20. (pentagon_ifs width height)) 100000;
    test_ifs (proba_ifs 30.)
      (proba_translate (f_width /. 2.) 20. (fern_proba width height)) 200000;
    test_ifs (proba_ifs 30.) (maple_proba width height) 200000;
    close_graph ();;
